/*
    --------------------------------------------------------------------
                             RMF RDS Widget
                Copyright (c) 2007-2012 Jacek Jędrzejewski
                    http://www.jacek.jedrzejewski.name/
                          All rights reserved.
    --------------------------------------------------------------------

    This file is part of RMF RDS Widget.
*/

package rds;

import flash.system.System;

import flash.events.Event;
import flash.events.ProgressEvent;
import flash.events.HTTPStatusEvent;
import flash.events.IOErrorEvent;
import flash.media.Sound;
import flash.media.SoundChannel;
import flash.media.SoundTransform;
import flash.media.SoundLoaderContext;
import flash.media.ID3Info;
import flash.net.URLRequest;

import flash.external.ExternalInterface;

import rds.GlobalTimer;

typedef Options = {
    var titleChangeCallback:String;
    var memoryMonitorCallback:String;
    var bufferingCallback:String;
    var errorCallback:String;
    var debugTraceCallback:String;
    
    var streamAddress:String;
    var volume:Float;
    var maxMemory:Int;
    var debug:Bool;
}

class Player
{
    public var sound:Sound;
    public var transform:SoundTransform;
    public var context:SoundLoaderContext;
    public var channel:SoundChannel;
    
    public var config:Options;
    
    public static var instance:Player;
    
    private var reconnect:Int;
    
    private var complete:Bool;
    
    public function new():Void
    {
        Player.instance = this;
        
        this.reconnect = 0;
        this.complete = false;
        
        this.config = {
            titleChangeCallback: '',
            memoryMonitorCallback: '',
            bufferingCallback: '',
            errorCallback: '',
            debugTraceCallback: '',
            
            streamAddress: '',
            volume: 0.5,
            maxMemory: 50,
            debug: false
        };

        try
        {
            this.transform = new SoundTransform();
            this.context = new SoundLoaderContext(3*1000);
            
            GlobalTimer.setInterval(function(){
                if(Player.instance.config.memoryMonitorCallback != '')
                {
                    ExternalInterface.call(Player.instance.config.memoryMonitorCallback, System.totalMemory);
                }
                
                if(Std.int(System.totalMemory) > Std.int(1024*1024*Player.instance.config.maxMemory))
                {
                    Player.instance.cleanMemory();
                }
            }, 1000, []);
            
            GlobalTimer.setInterval(function(){
                if(Player.instance.sound != null)
                {
                    if(Player.instance.config.bufferingCallback != '')
                    {
                        ExternalInterface.call(Player.instance.config.bufferingCallback, Player.instance.sound.isBuffering);
                    }
                }
            }, 100, []);
        }
        catch(msg:String)
        {
            trace("new() - Error message: "+msg);
        }
        catch(errorCode:Int)
        {
            trace("new() - Error #"+errorCode);
        }
        catch(unknown:Dynamic)
        {
            trace("new() - Unknown exception: "+Std.string(unknown));
        }
    }
    
    public function stop():Void
    {
        if(Player.instance.config.debug)
        {
            trace('STOP STREAM');
        }
        
        try
        {
            try
            {
                if(this.channel != null)
                {
                    this.channel.stop();
                }
                
                this.channel = null;
                    
                if(this.sound != null)
                {
                    this.sound.close();
                }
                
                this.sound = null;
            }
            catch(unknown:Dynamic)
            { }
            
            this.complete = false;
        }
        catch(msg:String)
        {
            trace("stop() - Error message: "+msg);
        }
        catch(errorCode:Int)
        {
            trace("stop() - Error #"+errorCode);
        }
        catch(unknown:Dynamic)
        {
            trace("stop() - Unknown exception: "+Std.string(unknown));
        }
    }
    
    public function play():Void
    {
        if(Player.instance.config.debug)
        {
            trace('PLAY STREAM: '+this.getAddress());
        }
        this.complete = false;
        
        try
        {
            try
            {
                if(this.channel != null)
                {
                    this.channel.stop();
                }
                
                this.channel = null;
                
                if(this.sound != null)
                {
                    this.sound.close();
                }
                
                this.sound = null;
            }
            catch(unknown:Dynamic)
            { }
            
            this.sound = new Sound();
            this.sound.addEventListener(Event.ID3, this.titleChangeHandler);
            this.sound.addEventListener(IOErrorEvent.IO_ERROR, this.errorHandler);
            
            this.sound.load(new URLRequest(this.getAddress()), this.context);
            
            this.changeVolume(this.config.volume);
            
            this.complete = true;
            
            this.channel = this.sound.play(0, 0, transform);
        }
        catch(msg:String)
        {
            trace("play() - Error message: "+msg);
        }
        catch(errorCode:Int)
        {
            trace("play() - Error #"+errorCode);
        }
        catch(unknown:Dynamic)
        {
            trace("play() - Unknown exception: "+Std.string(unknown));
        }
    }
    
    public function changeVolume(volume:Float):Dynamic
    {
        if(Player.instance.config.debug)
        {
            trace('VOLUME: '+volume);
        }
        
        this.config.volume = volume;
        
        this.transform.volume = volume;
        
        if(this.channel != null)
        {
            this.channel.soundTransform = this.transform;
        }
        return this.transform.volume;
    }
    
    public function setStream(address:String):Void
    {
        if(Player.instance.config.debug)
        {
            trace('SET STREAM: '+address);
        }
        
        this.stop();
        
        this.config.streamAddress = address;
    }
    
    public function configure(config:Dynamic):Void
    {
        if(Player.instance.config.debug)
        {
            trace(config);
        }

        for(field in Reflect.fields(this.config))
        {
            if(Reflect.hasField(config, field))
            {
                Reflect.setField(this.config, field, Reflect.field(config, field));
            }
            
            if(Player.instance.config.debug)
            {
                trace(field+" > "+Reflect.field(this.config, field));
            }
        }
    }
    
    public function cleanMemory():Void
    {
        //if(Player.instance.config.debug)
        //{
            trace('CLEAN MEMORY');
        //}
        
        try
        {
            this.stop();
            this.play();
        }
        catch(msg:String)
        {
            trace("cleanMemory() - Error message: "+msg);
        }
        catch(errorCode:Int)
        {
            trace("cleanMemory() - Error #"+errorCode);
        }
        catch(unknown:Dynamic)
        {
            trace("cleanMemory() - Unknown exception: "+Std.string(unknown));
        }
    }
    
    private function getAddress():String
    {
        return this.config.streamAddress+'?'+Math.random();
    }
    
    private function titleChangeHandler(e:Event):Void
    {
        var info:ID3Info = sound.id3;
        
        if(Player.instance.config.debug)
        {
            trace('TITLE: '+info.artist+' - '+info.songName);
        }
        
        if(Player.instance.config.memoryMonitorCallback != '')
        {
            ExternalInterface.call(Player.instance.config.titleChangeCallback, info);
        }
    }
    
    private function errorHandler(e:IOErrorEvent):Void
    {
        if(Player.instance.config.debug)
        {
            trace('ERROR: '+e.text);
        }
        
        if(Player.instance.config.errorCallback != '')
        {
            ExternalInterface.call(Player.instance.config.errorCallback, e.text);
        }
    }
}